# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022 Salvador E. Tropea
# Copyright (c) 2020-2022 Instituto Nacional de Tecnología Industrial
# License: GPL-3.0
# Project: KiBot (formerly KiPlot)
from fnmatch import fnmatch
import os
import requests
import tempfile
from .misc import W_MISS3D, W_FAILDL, W_DOWN3D, DISABLE_3D_MODEL_TEXT
from .gs import GS
from .out_base import VariantOptions, BaseOutput
from .kicad.config import KiConf
from .macros import macros, document  # noqa: F401
from . import log

logger = log.get_logger()


def do_expand_env(fname, used_extra, extra_debug):
    # Is it using ALIAS:xxxxx?
    force_used_extra = False
    if ':' in fname:
        ind = fname.index(':')
        alias_name = fname[:ind]
        rest = fname[ind+1:]
        if alias_name in KiConf.aliases_3D:
            # Yes, replace the alias
            fname = os.path.join(KiConf.aliases_3D[alias_name], rest)
            # Make sure the name we created is what kicad2step gets
            force_used_extra = True
            if extra_debug:
                logger.debug("- Replaced alias {} -> {}".format(alias_name+':'+rest, fname))
    full_name = KiConf.expand_env(fname, used_extra, ref_dir=GS.pcb_dir)
    if extra_debug:
        logger.debug("- Expanded {} -> {}".format(fname, full_name))
    if os.path.isfile(full_name) or ':' not in fname or GS.global_disable_3d_alias_as_env:
        full_name_cwd = KiConf.expand_env(fname, used_extra, ref_dir=os.getcwd())
        if os.path.isfile(full_name_cwd):
            full_name = full_name_cwd
            force_used_extra = True
        if force_used_extra:
            used_extra[0] = True
        return full_name
    ind = fname.index(':')
    alias_name = fname[:ind]
    rest = fname[ind+1:]
    new_fname = '${'+alias_name+'}'+os.path.sep+rest
    new_full_name = KiConf.expand_env(new_fname, used_extra)
    if extra_debug:
        logger.debug("- Expanded {} -> {}".format(new_fname, new_full_name))
    if os.path.isfile(new_full_name):
        used_extra[0] = True
        return new_full_name
    return full_name


class Base3DOptions(VariantOptions):
    def __init__(self):
        with document:
            self.no_virtual = False
            """ *Used to exclude 3D models for components with 'virtual' attribute """
            self.download = True
            """ *Downloads missing 3D models from KiCad git. Only applies to models in KISYS3DMOD """
            self.kicad_3d_url = 'https://gitlab.com/kicad/libraries/kicad-packages3D/-/raw/master/'
            """ Base URL for the KiCad 3D models """
        # Temporal dir used to store the downloaded files
        self._tmp_dir = None
        super().__init__()
        self._expand_id = '3D'

    def download_model(self, url, fname, rel_dirs):
        """ Download the 3D model from the provided URL """
        logger.debug('Downloading `{}`'.format(url))
        failed = False
        try:
            r = requests.get(url, allow_redirects=True)
        except Exception:
            failed = True
        if failed or r.status_code != 200:
            logger.warning(W_FAILDL+'Failed to download `{}`'.format(url))
            return None
        if self._tmp_dir is None:
            self._tmp_dir = tempfile.mkdtemp()
            rel_dirs.append(self._tmp_dir)
            logger.debug('Using `{}` as temporal dir for downloaded files'.format(self._tmp_dir))
        dest = os.path.join(self._tmp_dir, fname)
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        with open(dest, 'wb') as f:
            f.write(r.content)
        return dest

    def download_models(self, rename_filter=None, rename_function=None, rename_data=None):
        """ Check we have the 3D models.
            Inform missing models.
            Try to download the missing models
            Stores changes in self.undo_3d_models_rep """
        models_replaced = False
        # Load KiCad configuration so we can expand the 3D models path
        KiConf.init(GS.pcb_file)
        # List of models we already downloaded
        downloaded = set()
        # For the mode where we copy the 3D models
        source_models = set()
        is_copy_mode = rename_filter is not None
        rel_dirs = getattr(rename_data, 'rel_dirs', [])
        extra_debug = GS.debug_level > 3
        # Look for all the footprints
        for m in GS.get_modules():
            ref = m.GetReference()
            # Extract the models (the iterator returns copies)
            models = m.Models()
            models_l = []
            while not models.empty():
                models_l.append(models.pop())
            # Look for all the 3D models for this footprint
            for m3d in models_l:
                if m3d.m_Filename.endswith(DISABLE_3D_MODEL_TEXT):
                    # Skip models we intentionally disabled using a bogus name
                    if extra_debug:
                        logger.debug("- Skipping {} (disabled)".format(m3d.m_Filename))
                    continue
                if is_copy_mode and not fnmatch(m3d.m_Filename, rename_filter):
                    # Skip filtered footprints
                    continue
                used_extra = [False]
                full_name = do_expand_env(m3d.m_Filename, used_extra, extra_debug)
                if not os.path.isfile(full_name):
                    logger.debugl(2, 'Missing 3D model file {} ({})'.format(full_name, m3d.m_Filename))
                    # Missing 3D model
                    if self.download and (m3d.m_Filename.startswith('${KISYS3DMOD}/') or
                                          m3d.m_Filename.startswith('${KICAD6_3DMODEL_DIR}/')):
                        # This is a model from KiCad, try to download it
                        fname = m3d.m_Filename[m3d.m_Filename.find('/')+1:]
                        replace = None
                        if full_name in downloaded:
                            # Already downloaded
                            replace = os.path.join(self._tmp_dir, fname)
                        else:
                            # Download the model
                            url = self.kicad_3d_url+fname
                            replace = self.download_model(url, fname, rel_dirs)
                            if replace:
                                # Successfully downloaded
                                downloaded.add(full_name)
                                # If this is a .wrl also download the .step
                                if url.endswith('.wrl'):
                                    url = url[:-4]+'.step'
                                    fname = fname[:-4]+'.step'
                                    self.download_model(url, fname, rel_dirs)
                        if replace:
                            source_models.add(replace)
                            old_name = m3d.m_Filename
                            new_name = replace if not is_copy_mode else rename_function(rename_data, replace)
                            self.undo_3d_models[new_name] = old_name
                            m3d.m_Filename = new_name
                            models_replaced = True
                    if full_name not in downloaded:
                        logger.warning(W_MISS3D+'Missing 3D model for {}: `{}`'.format(ref, full_name))
                else:  # File was found
                    if used_extra[0] or is_copy_mode:
                        # The file is there, but we got it expanding a user defined text
                        # This is completely valid for KiCad, but kicad2step doesn't support it
                        source_models.add(full_name)
                        old_name = m3d.m_Filename
                        new_name = full_name if not is_copy_mode else rename_function(rename_data, full_name)
                        self.undo_3d_models[new_name] = old_name
                        m3d.m_Filename = new_name
                        if not models_replaced and extra_debug:
                            logger.debug('- Modifying models with text vars')
                        models_replaced = True
            # Push the models back
            for model in models_l:
                models.push_front(model)
        if downloaded:
            logger.warning(W_DOWN3D+' {} 3D models downloaded'.format(len(downloaded)))
        return models_replaced if not is_copy_mode else list(source_models)

    def list_models(self):
        """ Get the list of 3D models """
        # Load KiCad configuration so we can expand the 3D models path
        KiConf.init(GS.pcb_file)
        models = set()
        # Look for all the footprints
        for m in GS.get_modules():
            # Look for all the 3D models for this footprint
            for m3d in m.Models():
                full_name = KiConf.expand_env(m3d.m_Filename)
                if os.path.isfile(full_name):
                    models.add(full_name)
        return list(models)

    def filter_components(self):
        if not self._comps:
            # No variant/filter to apply
            if self.download_models():
                # Some missing components found and we downloaded them
                # Save the fixed board
                ret = self.save_tmp_board()
                # Undo the changes done during download
                self.undo_3d_models_rename(GS.board)
                return ret
            return GS.pcb_file
        self.filter_pcb_components(GS.board, do_3D=True, do_2D=True)
        self.download_models()
        fname = self.save_tmp_board()
        self.unfilter_pcb_components(GS.board, do_3D=True, do_2D=True)
        return fname

    def get_targets(self, out_dir):
        return [self._parent.expand_filename(out_dir, self.output)]


class Base3D(BaseOutput):
    def __init__(self):
        super().__init__()

    def get_dependencies(self):
        files = super().get_dependencies()
        files.extend(self.options.list_models())
        return files



# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022 Salvador E. Tropea
# Copyright (c) 2020-2022 Instituto Nacional de Tecnología Industrial
# License: GPL-3.0
# Project: KiBot (formerly KiPlot)
"""
Dependencies:
  - from: KiAuto
    role: mandatory
    command: eeschema_do
    version: 1.5.4
"""
import os
from sys import exit
import xml.etree.ElementTree as ET
from .macros import macros, document, pre_class  # noqa: F401
from .error import KiPlotConfigurationError
from .gs import GS
from .kiplot import exec_with_retry, add_extra_options, load_board
from .misc import BOM_ERROR, NETLIST_DIFF, W_PARITY, MISSING_TOOL
from .log import get_logger
from .optionable import Optionable
import pcbnew

logger = get_logger(__name__)


class Update_XMLOptions(Optionable):
    """ Reference sorting options """
    def __init__(self):
        super().__init__()
        with document:
            self.enabled = True
            """ Enable the update. This is the replacement for the boolean value """
            self.check_pcb_parity = False
            """ *Check if the PCB and Schematic are synchronized.
                This is equivalent to the *Test for parity between PCB and schematic* of the DRC dialog.
                Only available for KiCad 6 """
            self.as_warnings = False
            """ Inform the problems as warnings and don't stop """


@pre_class
class Update_XML(BasePreFlight):  # noqa: F821
    """ [boolean=false|dict] Update the XML version of the BoM (Bill of Materials).
        To ensure our generated BoM is up to date.
        Note that this isn't needed when using the internal BoM generator (`bom`).
        You can compare the PCB and schematic netlists using it """
    def __init__(self, name, value):
        super().__init__(name, value)
        self._check_pcb_parity = False
        if isinstance(value, bool):
            self._enabled = value
        elif isinstance(value, dict):
            f = Update_XMLOptions()
            f.set_tree(value)
            f.config(self)
            self._enabled = f.enabled
            self._check_pcb_parity = f.check_pcb_parity
            self.options = f
            self._pcb_related = True
        else:
            raise KiPlotConfigurationError('must be boolean or dict')
        self._sch_related = True

    @classmethod
    def get_doc(cls):
        return cls.__doc__, Update_XMLOptions

    def get_targets(self):
        """ Returns a list of targets generated by this preflight """
        return [GS.sch_no_ext+'.xml']

    def check_components(self, comps, errors):
        found_comps = set()
        for m in GS.get_modules():
            ref = m.GetReference()
            found_comps.add(ref)
            if ref not in comps:
                errors.append('{} found in PCB, but not in schematic'.format(ref))
                continue
            sch_fp = comps[ref]
            pcb_fp = m.GetFPIDAsString()
            if sch_fp != pcb_fp:
                errors.append('{} footprint mismatch (PCB: {} vs schematic: {})'.format(ref, pcb_fp, sch_fp))
        for ref in set(comps.keys()).difference(found_comps):
            errors.append('{} found in schematic, but not in PCB'.format(ref))

    def check_nets(self, net_names, net_nodes, errors):
        # Total count
        con = GS.board.GetConnectivity()
        pcb_net_count = con.GetNetCount()-1  # Removing the bogus net 0
        sch_net_count = len(net_names)
        if pcb_net_count != sch_net_count:
            errors.append('Net count mismatch (PCB {} vs schematic {})'.format(pcb_net_count, sch_net_count))
        net_info = GS.board.GetNetInfo()
        # Names and connection
        for n in net_info.NetsByNetcode():
            if not n:
                # Bogus net code 0
                continue
            if n not in net_names:
                errors.append('PCB net code {} not in schematic'.format(n))
                continue
            net = net_info.GetNetItem(n)
            net_name = net.GetNetname()
            sch_name = net_names[n]
            if net_name != sch_name:
                errors.append('PCB net code {} name mismatch ({} vs {})'.format(n, net_name, sch_name))
            sch_nodes = net_nodes[n]
            pcb_nodes = {pad.GetParent().GetReference()+' pin '+pad.GetNumber()
                         for pad in con.GetNetItems(n, [pcbnew.PCB_PAD_T])}
            dif = pcb_nodes-sch_nodes
            if dif:
                errors.append('PCB net code {} extra connection/s: {}'.format(n, ','.join(list(dif))))
            dif = sch_nodes-pcb_nodes
            if dif:
                errors.append('PCB net code {} missing connection/s: {}'.format(n, ','.join(list(dif))))

    def check_pcb_parity(self):
        if GS.ki5:
            logger.error('PCB vs schematic parity only available for KiCad 6')
            exit(MISSING_TOOL)
        fname = GS.sch_no_ext+'.xml'
        logger.debug('Loading XML: '+fname)
        try:
            tree = ET.parse(fname)
        except Exception as e:
            raise KiPlotConfigurationError('Errors parsing {}\n{}'.format(fname, e))
        root = tree.getroot()
        if root.tag != 'export':
            raise KiPlotConfigurationError("{} isn't a valid netlist".format(fname))
        # Check version? root.attrib.get('version')
        components = root.find('components')
        comps = {}
        if components is not None:
            for c in components.iter('comp'):
                ref = c.attrib.get('ref')
                fp = c.find('footprint')
                fp = fp.text if fp is not None else ''
                logger.debugl(2, '- {}: {}'.format(ref, fp))
                comps[ref] = fp
        netlist = root.find('nets')
        net_names = {}
        net_nodes = {}
        if netlist is not None:
            for n in netlist.iter('net'):
                code = int(n.get('code'))
                net_names[code] = n.get('name')
                net_nodes[code] = {node.get('ref')+' pin '+node.get('pin') for node in n.iter('node')}
        # Check with the PCB
        errors = []
        load_board()
        # Check components
        self.check_components(comps, errors)
        # Check the nets
        self.check_nets(net_names, net_nodes, errors)
        # Report errors
        if errors:
            if self.options.as_warnings:
                for e in errors:
                    logger.warning(W_PARITY+e)
            else:
                for e in errors:
                    logger.error(e)
                exit(NETLIST_DIFF)

    def run(self):
        command = self.ensure_tool('KiAuto')
        out_dir = self.expand_dirname(GS.out_dir)
        cmd = [command, 'bom_xml', GS.sch_file, out_dir]
        # If we are in verbose mode enable debug in the child
        cmd, video_remove = add_extra_options(cmd)
        # While creating the XML we run a BoM plug-in that creates a useless BoM
        # We remove it, unless this is already there
        side_effect_file = os.path.join(out_dir, GS.sch_basename+'.csv')
        remove_side_effect_file = not os.path.isfile(side_effect_file)
        logger.info('- Updating BoM in XML format')
        ret = exec_with_retry(cmd)
        if remove_side_effect_file and os.path.isfile(side_effect_file):
            os.remove(side_effect_file)
        if ret:
            logger.error('Failed to update the BoM, error %d', ret)
            exit(BOM_ERROR)
        if video_remove:
            video_name = os.path.join(self.expand_dirname(GS.out_dir), 'bom_xml_eeschema_screencast.ogv')
            if os.path.isfile(video_name):
                os.remove(video_name)
        if self._check_pcb_parity:
            self.check_pcb_parity()



# -*- coding: utf-8 -*-
# Copyright (c) 2022 Salvador E. Tropea
# Copyright (c) 2022 Instituto Nacional de Tecnología Industrial
# License: GPL-3.0
# Project: KiBot (formerly KiPlot)
import fnmatch
import glob
import os
import re
from shutil import copy2, rmtree
from sys import exit
from .error import KiPlotConfigurationError
from .gs import GS
from .kiplot import config_output, get_output_dir, run_output
from .kicad.config import KiConf
from .misc import WRONG_ARGUMENTS, INTERNAL_ERROR, W_COPYOVER
from .optionable import Optionable
from .out_base_3d import Base3DOptions
from .registrable import RegOutput
from .macros import macros, document, output_class  # noqa: F401
from . import log

logger = log.get_logger()


def may_be_rel(file):
    rel_file = os.path.relpath(file)
    if len(rel_file) < len(file):
        return rel_file
    return file


class FilesList(Optionable):
    def __init__(self):
        super().__init__()
        with document:
            self.source = '*'
            """ *File names to add, wildcards allowed. Use ** for recursive match.
                By default this pattern is applied to the current working dir.
                See the `from_outdir` option """
            self.source_type = 'files'
            """ *[files,out_files,output,3d_models] How to interpret `source`.
                `files`: is a pattern for files relative to the working directory.
                `out_files`: is a pattern for files relative to output dir specified
                with `-d` command line option.
                `output`: is the name of an `output`.
                `3d_models`: is a pattern to match the name of the 3D models extracted
                from the PCB. """
            self.filter = '.*'
            """ A regular expression that source files must match """
            self.dest = ''
            """ Destination directory inside the output dir, empty means the same of the file
                relative to the source directory.
                For the `3d_models` type you can use DIR+ to create subdirs under DIR """
            self.save_pcb = False
            """ Only usable for the `3d_models` mode.
                Save a PCB copy modified to use the copied 3D models """

    def apply_rename(self, fname):
        is_abs = os.path.isabs(fname)
        append_mode = self.dest and self.dest[-1] == '+'
        if self.dest and not append_mode:
            # A destination specified by the user
            dest = os.path.basename(fname)
        elif is_abs:
            for d in self.rel_dirs:
                if d is not None and fname.startswith(d):
                    dest = os.path.relpath(fname, d)
                    break
            else:
                dest = os.path.basename(fname)
        else:
            dest = os.path.relpath(fname, os.getcwd())
        return '${KIPRJMOD}/'+os.path.join(self.output_dir, dest)


class Copy_FilesOptions(Base3DOptions):
    def __init__(self):
        with document:
            self.files = FilesList
            """ *[list(dict)] Which files will be included """
            self.follow_links = True
            """ Store the file pointed by symlinks, not the symlink """
            self.link_no_copy = False
            """ Create symlinks instead of copying files """
        super().__init__()
        self._expand_id = 'copy'
        self._expand_ext = 'files'

    def config(self, parent):
        super().config(parent)
        if isinstance(self.files, type):
            KiPlotConfigurationError('No files provided')

    def get_from_output(self, f, no_out_run):
        from_output = f.source
        logger.debugl(2, '- From output `{}`'.format(from_output))
        out = RegOutput.get_output(from_output)
        if out is not None:
            config_output(out)
            out_dir = get_output_dir(out.dir, out, dry=True)
            files_list = out.get_targets(out_dir)
            logger.debugl(2, '- List of files: {}'.format(files_list))
        else:
            logger.error('Unknown output `{}` selected in {}'.format(from_output, self._parent))
            exit(WRONG_ARGUMENTS)
        # Check if we must run the output to create the files
        if not no_out_run:
            for file in files_list:
                if not os.path.isfile(file):
                    # The target doesn't exist
                    if not out._done:
                        # The output wasn't created in this run, try running it
                        run_output(out)
                    if not os.path.isfile(file):
                        # Still missing, something is wrong
                        logger.error('Unable to generate `{}` from {}'.format(file, out))
                        exit(INTERNAL_ERROR)
        return files_list

    def get_3d_models(self, f):
        """ Look for the 3D models and make a list, optionally download them """
        GS.check_pcb()
        dest_dir = f.dest
        if dest_dir and dest_dir[-1] == '+':
            dest_dir = dest_dir[:-1]
        f.output_dir = dest_dir
        # Apply any variant
        self.filter_pcb_components(GS.board, do_3D=True, do_2D=True)
        # Download missing models and rename all collect 3D models (renamed)
        f.rel_dirs = self.rel_dirs
        files_list = self.download_models(rename_filter=f.source, rename_function=FilesList.apply_rename, rename_data=f)
        if f.save_pcb:
            fname = os.path.join(self.output_dir, os.path.basename(GS.pcb_file))
            logger.debug('Saving the PCB to '+fname)
            GS.board.Save(fname)
            GS.copy_project(fname)
        if not self._comps:
            # We must undo the download/rename
            self.undo_3d_models_rename(GS.board)
        else:
            self.unfilter_pcb_components(GS.board, do_3D=True, do_2D=True)
        # Also include the step/wrl counterpart
        new_list = []
        for fn in files_list:
            if fn.endswith('.wrl'):
                fn = fn[:-4]+'.step'
                if os.path.isfile(fn):
                    new_list.append(fn)
            elif fn.endswith('.step'):
                fn = f[:-5]+'.wrl'
                if os.path.isfile(fn):
                    new_list.append(fn)
        return files_list+fnmatch.filter(new_list, f.source)

    def get_files(self, no_out_run=False):
        files = []
        src_dir_cwd = os.getcwd()
        src_dir_outdir = self.expand_filename_sch(GS.out_dir)
        self.rel_dirs = [os.path.normpath(os.path.join(GS.pcb_dir, KiConf.models_3d_dir)),
                         os.path.normpath(os.path.join(GS.pcb_dir, KiConf.party_3rd_dir)),
                         GS.pcb_dir]
        for f in self.files:
            from_outdir = False
            if f.source_type == 'out_files' or f.source_type == 'output':
                from_outdir = True
            src_dir = src_dir_outdir if from_outdir else src_dir_cwd
            mode_3d = f.source_type == '3d_models'
            mode_3d_append = mode_3d and f.dest and f.dest[-1] == '+'
            # Get the list of candidates
            files_list = None
            if f.source_type == 'output':
                files_list = self.get_from_output(f, no_out_run)
            elif mode_3d:
                files_list = self.get_3d_models(f)
            else:  # files and out_files
                source = f.expand_filename_both(f.source, make_safe=False)
                files_list = glob.iglob(os.path.join(src_dir, source), recursive=True)
                if GS.debug_level > 1:
                    files_list = list(files_list)
                    logger.debug('- Pattern {} list of files: {}'.format(source, files_list))
            # Filter and adapt them
            for fname in filter(re.compile(f.filter).match, files_list):
                fname_real = os.path.realpath(fname) if self.follow_links else os.path.abspath(fname)
                # Compute the destination directory
                dest = fname
                is_abs = os.path.isabs(fname)
                if f.dest and not mode_3d_append:
                    # A destination specified by the user
                    dest = os.path.join(f.dest, os.path.basename(fname))
                elif mode_3d and is_abs:
                    for d in self.rel_dirs:
                        if d is not None and fname.startswith(d):
                            dest = os.path.relpath(dest, d)
                            break
                    else:
                        logger.error(fname)
                        logger.error(self.rel_dirs)
                        dest = os.path.basename(fname)
                    if mode_3d_append:
                        dest = os.path.join(f.dest[:-1], dest)
                else:
                    dest = os.path.relpath(dest, src_dir)
                files.append((fname_real, dest))

        return files

    def get_targets(self, out_dir):
        files = self.get_files(no_out_run=True)
        return sorted([os.path.join(out_dir, v) for v in files.values()])

    def get_dependencies(self):
        files = self.get_files(no_out_run=True)
        return files.keys()

    def run(self, output):
        # Output file name
        logger.debug('Collecting files')
        # Collect the files
        files = self.get_files()
        logger.debug('Copying files')
        output += os.path.sep
        copied = {}
        for (src, dst) in files:
            dest = os.path.join(output, dst)
            dest_dir = os.path.dirname(dest)
            if not os.path.isdir(dest_dir):
                os.makedirs(dest_dir)
            logger.debug('- {} -> {}'.format(src, dest))
            if dest in copied:
                logger.warning(W_COPYOVER+'`{}` and `{}` both are copied to `{}`'.
                               format(may_be_rel(src), may_be_rel(copied[dest]), may_be_rel(dest)))
            if os.path.isfile(dest) or os.path.islink(dest):
                os.remove(dest)
            if self.link_no_copy:
                os.symlink(os.path.relpath(src, os.path.dirname(dest)), dest)
            else:
                copy2(src, dest)
            copied[dest] = src
        # Remove the downloaded 3D models
        if self._tmp_dir:
            rmtree(self._tmp_dir)
            self._tmp_dir = None


@output_class
class Copy_Files(BaseOutput):  # noqa: F821
    """ Files copier
        Used to copy files to the output directory.
        Useful when an external tool is used to compress the output directory.
        Note that you can use the `compress` output to create archives """
    def __init__(self):
        super().__init__()
        # Make it low priority so it gets created after all the other outputs
        self.priority = 11
        with document:
            self.options = Copy_FilesOptions
            """ *[dict] Options for the `copy_files` output """
        self._none_related = True
        # The help is inherited and already mentions the default priority
        self.fix_priority_help()

    def get_dependencies(self):
        return self.options.get_dependencies()

    def run(self, output_dir):
        # No output member, just a dir
        self.options.output_dir = output_dir
        self.options.run(output_dir)
